#include "MainApplication.h"
#include "FileLogger.h"
#include "MidiReader.h"
#include "File.h"
#include "DocumentConverter.h"
#include <filesystem>

MainApplication::MainApplication()
	 : mDatabaseManager(),
	   mCommandLineArgs()
{

}

MainApplication::~MainApplication()
{

}

void MainApplication::Initialize(CommandLineArgsUPtr commandLineArgs)
{
    mCommandLineArgs = std::move(commandLineArgs);
    std::string launch_path = mCommandLineArgs->GetLaunchPath().string();

	FileLogger::GetInstance().SetWorkDirectory(launch_path);
	FileLogger::GetInstance().Open();
	MLOG_INFO("Launched");

	mDatabaseManager = DatabaseManager::Create();
	mDatabaseManager->CreateDatabase(launch_path + "/database.db");

	mNetworkManager = NetworkManager::Create();

	mAudioManager = AudioManager::Create();
}

void MainApplication::Run()
{
    std::string program_type;
    if(mCommandLineArgs->GetNumberOfArgs() > 1)
    {
        program_type = mCommandLineArgs->GetArg(1);
    }

    std::string input_path;
    std::string output_path;
    for(unsigned idx=1; idx<mCommandLineArgs->GetNumberOfArgs(); idx++)
    {
        auto arg = mCommandLineArgs->GetArg(idx);
        if(arg == "-input" && mCommandLineArgs->GetNumberOfArgs() > idx+1)
        {
            input_path = mCommandLineArgs->GetArg(idx + 1);
        }
        else if(arg == "-output" && mCommandLineArgs->GetNumberOfArgs() > idx+1)
        {
            output_path = mCommandLineArgs->GetArg(idx + 1);
        }
    }

    if(program_type == "-server")
    {
        RunServer();
    }
    else if(program_type == "-audio")
    {
        PlayAudio();
    }
    else if(program_type == "-convert")
    {
        ConvertDocument(input_path, output_path);
    }
    else
    {
        MLOG_ERROR("Unknown program type: " + program_type);
    }
}

void MainApplication::RunServer()
{
	mNetworkManager->RunHttpServer();
}

void MainApplication::PlayAudio()
{
	//MidiReader reader;
	//reader.Read("/home/james/sample.mid");
    auto device = AudioDevice::Create();
    mAudioManager->GetAudioInterface()->OpenDevice(device);
    mAudioManager->GetAudioInterface()->Play(device);
}

void MainApplication::ConvertDocument(const std::string& inputPath, const std::string& outputPath)
{
    auto input_file = File(std::filesystem::path(inputPath));
    auto output_file = File(std::filesystem::path(outputPath));
    MLOG_INFO("Converting: " + inputPath + " to " + outputPath);
    DocumentConverter converter;
    converter.Convert(&input_file, &output_file);
}

void MainApplication::ShutDown()
{
	mDatabaseManager->OnShutDown();
	mNetworkManager->ShutDown();
	MLOG_INFO("Shut down");
	FileLogger::GetInstance().Close();
}

std::shared_ptr<MainApplication> MainApplication::Create()
{
	return std::make_shared<MainApplication>();
}
